/*
 * File         : BV_VideoApp.C
 * Author       : Tomasz Stachowiak
 * Created      : Fri Oct 16 17:16:04 1998
 * Revision     : $Revision$
 * Updated      : $Date$
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include <sys/types.h>
#include <sys/time.h>
#include <semaphore.h>

#include "H263.h"
#include "H261.h"
#include "rtp.h"
#include "MainForm.h"
#include "BV.h"
#include "BV_Application.h"
#include "BV_VideoCapThread.h"
#include "BV_Member.h"
#include "BV_Session_Control.h"
#include "BV_VideoApp.h"

FILE *log;

BV_VideoApp::BV_VideoApp() : BV_Application(VIDEO_PLUGIN_PORT)
{
  session_ = BV_Session_Control::instance();
  main_form_ = session_->getMainForm();

  state_ = STOP;
  compression_ = H263;
  first = 1;
  frame_no = 0;
  msg_no = 0;
  kbps_limit_ = 20;
  wait_time_ = 10;
  fps_max_ = 5;

  sem_init(&semH263_, 0, 1);
  sem_init(&semH261_, 0, 1);
  sem_init(&sem_, 0, 1);

  H263initEncoder(176, 144);
  H263initDecoder();

  /* H.261 */ 
  init_codec();
  init_encoder();
  init_decoder(); 

  initcapture(176, 144);
  init_global_dpy();

  log = fopen("video.log", "w");
  
  cap_thread_ = new BV_VideoCapThread(this);
  cap_thread_->start();
  
}

BV_VideoApp::~BV_VideoApp()
{
  cap_thread_->kill();
  //cap_thread_->stop();
  sleep(1);
  delete cap_thread_;
  fclose(log);
  session_->destroy();
}

int BV_VideoApp::process_message(RTPheader *header, char *body, int len)
{
  //char name[50];

  if(header->type != 2)
    {
      return -1;
    }

  int sender = header->ssrc;

  session_->lock();
  BV_Member *mbr = session_->get_member(sender);
  if(mbr == NULL)
    {
      session_->unlock();
      return -1;
    }
  
  mbr->update_video_stat(header, len);

  session_->unlock();

  //mbr->get_name(name);
  //session_->update_video_stat(mbr);

  switch (header->pt)
    {
    case H263:
      displayH263(mbr, body, len);
      break;

    case H261:
      displayH261(mbr, body, len);
      break;

    case YUV9:
      displayYUV9(mbr, body, len);
      break;

    case videoTagPlay:
      //session_->lock();
      //mbr->open_display();
      mbr->open_dpy_req();
      //session_->unlock();
      break;

    case videoTagStop:
      session_->lock();
      mbr->close_dpy_req();
      //mbr->close_display();
      session_->unlock();
      break;

    default:
      break;
    }

  return 0;
}


int BV_VideoApp::displayH263(BV_Member *mbr, char *data, int len)
{
  VideoPluginFrame *frame;
  struct timeval start, stop;
  unsigned char *dec_frame;
  int lastDecodeTime;
  int size, id;
  
  id = mbr->get_internal_id();

  gettimeofday(&start);
  frame = (VideoPluginFrame *)data;
  size = len - sizeof(VideoPluginFrame) + 1;

      session_->lock();
  sem_wait(&semH263_);
  dec_frame = (unsigned char *)H263decodeFrame(id, frame->data, size);
  sem_post(&semH263_);

  if(dec_frame != NULL)
    {
      //mbr->display((char *)dec_frame);
  //    session_->lock();
  mbr->display_req((char *)dec_frame);
  //    session_->unlock();
      free(dec_frame);
    }

      session_->unlock();
  gettimeofday(&stop);
  lastDecodeTime = (stop.tv_sec - start.tv_sec)*1000 
    + (stop.tv_usec - start.tv_usec)/1000;
  
  if(dec_frame == NULL)
    return -1;

  return 0;
}

int BV_VideoApp::displayH261(BV_Member *mbr, char *data, int len)
{
  VideoPluginFrame *frame;
  struct timeval start, stop;
  unsigned char *dec_frame;
  int lastDecodeTime;
  int size, id;
  
  id = mbr->get_internal_id();

  gettimeofday(&start);
  frame = (VideoPluginFrame *)data;
  size = len - sizeof(VideoPluginFrame) + 1;

  /*
  if(h261first)
    {
      decode_frame(frame->data, size, dec_frame, 1, mbr);
      h261first = 0;
    }
    else */
  sem_wait(&semH263_);
  dec_frame = (unsigned char *)malloc(176*144*3/2);
  if(dec_frame != NULL)
    {
      decode_frame(frame->data, size, (char *)dec_frame, 0, id);
    }
  sem_post(&semH263_);

  if(dec_frame != NULL)
    {
      session_->lock();
      mbr->display_req((char *)dec_frame);
      session_->unlock();
      free(dec_frame);
    }

  gettimeofday(&stop);
  lastDecodeTime = (stop.tv_sec - start.tv_sec)*1000 
    + (stop.tv_usec - start.tv_usec)/1000;
  
  if(dec_frame == NULL)
    return -1;

  return 0;
}

int BV_VideoApp::displayYUV9(BV_Member *mbr, char *data, int len)
{
  VideoPluginFrame *frame;
  struct timeval start, stop;
  unsigned char *dec_frame;
  int lastDecodeTime;
  int size;
  
  gettimeofday(&start);

  frame = (VideoPluginFrame *)data;
  size = len - sizeof(VideoPluginFrame) + 1;
  dec_frame = (unsigned char *)malloc(176*144*3/2);
  YVU9toYUV411((unsigned char *)frame->data, dec_frame, &dec_frame[176*144], 
	       &dec_frame[176*144*5/4], 176, 144);

  //mbr->display((char *)dec_frame);
  session_->lock();
  mbr->display_req((char *)dec_frame);
  session_->unlock();
  
  free(dec_frame);

  gettimeofday(&stop);
  lastDecodeTime = (stop.tv_sec - start.tv_sec)*1000 
    + (stop.tv_usec - start.tv_usec)/1000;
  
  if(dec_frame == NULL)
    return -1;

  return 0;
}



int BV_VideoApp::send_video()
{
  unsigned char *frame;
  char *buf;
  int size;

  if(state_ != PLAY)
    {
      sleep(1);
      return 0;
    }

  session_->lock();
  BV_Member *mbr = session_->get_member(session_->get_local_id());
  if(mbr == NULL)
    {
      session_->unlock();
      sleep(1);
      return -1;
    }
  session_->unlock();

  buf = (char *)malloc(10000);
  if(buf == NULL)
    {
      printf("Memory allocation error!\n");
      fflush(stdout);
      sleep(1);
      return -1;
    }

  switch(compression_)
    {
    case H263:
      frame = getFrame(176, 144);
      //mbr->display((char *)frame);
      session_->lock();
      mbr->display_req((char *)frame);
      session_->unlock();
      
      sem_wait(&semH263_);
      if(frame_no == 0)
	{
	  size = H263codeIntraFrame(frame, buf);
	}
      else
	{		
	  size = H263codeFrame(&frame, buf);
	}
      sem_post(&semH263_);

      free(frame);
      send_video_data(compression_, buf, size);
      break;
      
    case H261:
      frame = getFrame(176, 144);
      
      session_->lock();
      //mbr->display((char *)frame);
      mbr->display_req((char *)frame);
      session_->unlock();
      
      sem_wait(&semH263_);
      if(frame_no == 0)
	size = p64EncodeIntraFrame((char *)frame, buf);
      else
	size = p64EncodeFrame((char *)frame, buf);
      
      sem_post(&semH263_);

      free(frame);
      send_video_data(compression_, buf, size);
      
      break;
    }
  
  free(buf);

  RTPheader header;
  header.seq = msg_no;
  
  session_->lock();
  mbr->update_video_stat(&header, size);
  session_->unlock();

  //session_->update_video_stat(mbr);
  
  frame_no = (frame_no + 1) % INTRA_GAP;
  
  wait_for_next_frame(mbr);

  return 0;
}

void BV_VideoApp::play()
{
  //session_->lock();
  if(state_ == STOP)
    {
      BV_Member *mbr = session_->get_member(session_->get_local_id());
      if(mbr != NULL)
      	{
	  //mbr->open_display();
	  mbr->open_dpy_req();
	  state_ = PLAY;
	}
    } 
  //session_->unlock();
}

void BV_VideoApp::stop()
{
  session_->lock();
  if(state_ == PLAY)
    {
      state_ = STOP;
      BV_Member *mbr = session_->get_member(session_->get_local_id());
      if(mbr != NULL)
	{
	  mbr->close_dpy_req();
	  //mbr->close_display();
	}
    }
  session_->unlock();
}

int BV_VideoApp::send_video_data(int tag, char *data, int len)
{
  VideoPluginFrame *frame;
  int err;

  frame = (VideoPluginFrame *)malloc(sizeof(VideoPluginFrame) + len - 1);
   if (!frame)
   {
     fprintf(stderr, "Message allocation error\n");
     return -1;
   }

  if (len <= 0)
    return -2;

  msg_no = (msg_no + 1) % 256;
  frame->num = (unsigned char)msg_no;
  gettimeofday(&frame->time);
  memcpy(frame->data, data, len);

  err = send_data(compression_, (char *)frame, sizeof(VideoPluginFrame)+len-1);
  
  free(frame);

  return err;
}

void BV_VideoApp::update_options(Options *options)
{
  int tmp_state = state_;
  state_ = STOP;
  if(tmp_state == PLAY)
    sleep(1);

  first = 1;
  frame_no = 0;

  switch(options->connection)
    {
    case _modem28:
      if(compression_ != H263)
	{
	  options->fps_max = 10;
	}
      compression_ = H263;
      kbps_limit_ = 20;
      break;

    case _isdn:
      if(compression_ != H261)
	{
	  options->fps_max = 5;
	}
      compression_ = H261;
      kbps_limit_ = 60;
      break;
      
    case _isdn2:
      if(compression_ != H261)
	{
	  options->fps_max = 10;
	}
      compression_ = H261;
      kbps_limit_ = 120;
      break;

    case _lan:
      if(compression_ != H263)
	{
	  options->fps_max = 10;
	}
      compression_ = H261;
      kbps_limit_ = 1500;
      break;

    default:
      compression_ = H263;
      break;
    }

  fps_max_ = options->fps_max;

  state_ = tmp_state;
}

void BV_VideoApp::wait_for_next_frame(BV_Member *mbr)
{
  struct timeval tm;

  double kbps = 0;
  double fps = 0;

  if(mbr->get_video_data() != NULL)
    {
      kbps = mbr->get_video_data()->getKbps();
      fps = mbr->get_video_data()->getFps();
    }

  if(fps > fps_max_)
    wait_time_ *= 1.1;
  else
    {
      if(kbps > kbps_limit_)
	wait_time_ *= 1.1;

      if(kbps < kbps_limit_ * 0.8)
	wait_time_ *= 0.9;
    }

  if(wait_time_ < 1)
    wait_time_ = 1;

  tm.tv_sec = wait_time_ / 1000;
  tm.tv_usec = ((int)wait_time_ % 1000) * 1000;
  select(0, 0, 0, 0, &tm);
}
